#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <stdint.h>
#include <sys/time.h>

//#include <xmmintrin.h>    // – SSE
//#include <emmintrin.h>  // – SSE2
//#include <pmmintrin.h>  // – SSE3
#include <tmmintrin.h>    // – SSSE3
#include <smmintrin.h>    // – SSE4.1
//#include <ammintrin.h>  // – SSE4a
//#include <bmmintrin.h>  // – SSE5

#define ARRAYSIZE 100000000 

#ifndef MODE
	#define MODE 0
#endif

int main (int argc, char **argv) {
  
  struct timeval start, end;
  long time1;
  
  uint8_t *numberarray = malloc(ARRAYSIZE * sizeof(*numberarray));
  uint32_t i;

  if(numberarray == NULL) {
    fprintf (stderr, "error during memory allocation\n");
    exit(EXIT_FAILURE);
  }

  for (i = 0; i < ARRAYSIZE; i++) {
    numberarray[i] = rand() % 100;
  }

#if( MODE == 1 )
	
	/* Parallel count : 4 values at a time */
  
  gettimeofday (&start, NULL);

  __m128i next4numbers;
  __m128i next16numbers;

  const __m128i comparatorMask = _mm_set_epi32(42,42,42,42);
	__m128i accumulator          = _mm_setzero_si128();

  const __m128i shuffmask1 = _mm_set_epi8 (0xff, 0xff, 0xff, 3,
                                           0xff, 0xff, 0xff, 2,
                                           0xff, 0xff, 0xff, 1,
                                           0xff, 0xff, 0xff, 0);

  const __m128i shuffmask2 = _mm_set_epi8 (0xff, 0xff, 0xff, 7,
                                           0xff, 0xff, 0xff, 6,
                                           0xff, 0xff, 0xff, 5,
                                           0xff, 0xff, 0xff, 4);

  const __m128i shuffmask3 = _mm_set_epi8 (0xff, 0xff, 0xff, 11,
                                           0xff, 0xff, 0xff, 10,
                                           0xff, 0xff, 0xff, 9,
                                           0xff, 0xff, 0xff, 8);

  const __m128i shuffmask4 = _mm_set_epi8 (0xff, 0xff, 0xff, 15,
                                           0xff, 0xff, 0xff, 14,
                                           0xff, 0xff, 0xff, 13,
                                           0xff, 0xff, 0xff, 12);

  for (i = 0; i < ARRAYSIZE/16; i++) {
    next16numbers = _mm_loadu_si128((__m128i *) (numberarray+i*16));

    next4numbers = _mm_shuffle_epi8(next16numbers, shuffmask1);
    next4numbers = _mm_cmpgt_epi32(next4numbers, comparatorMask);
    accumulator  = _mm_add_epi32(accumulator, next4numbers);

    next4numbers = _mm_shuffle_epi8(next16numbers, shuffmask2);
    next4numbers = _mm_cmpgt_epi32(next4numbers, comparatorMask);
    accumulator  = _mm_add_epi32(accumulator, next4numbers);

    next4numbers = _mm_shuffle_epi8(next16numbers, shuffmask3);
    next4numbers = _mm_cmpgt_epi32(next4numbers, comparatorMask);
    accumulator  = _mm_add_epi32(accumulator, next4numbers);

    next4numbers = _mm_shuffle_epi8(next16numbers, shuffmask4);
    next4numbers = _mm_cmpgt_epi32(next4numbers, comparatorMask);
    accumulator  = _mm_add_epi32(accumulator, next4numbers);
  }

  uint32_t parallelcnt = _mm_extract_epi32(accumulator,0)+_mm_extract_epi32(accumulator,1)+_mm_extract_epi32(accumulator,2)+_mm_extract_epi32(accumulator,3);
  parallelcnt *= -1;

  gettimeofday (&end, NULL);

  printf("Paralle-4 : count = %d\n",parallelcnt);

#elif( MODE == 2 )

	/* Parallel count : 16 values at a time - version 1 */
  
  uint32_t j;

  gettimeofday (&start, NULL);
  
  uint32_t parallelcnt = 0;
  __m128i next16numbers;
  const __m128i comparatorMask = _mm_set_epi8(42,42,42,42,42,42,42,42,42,42,42,42,42,42,42,42);
  __m128i accumulator          = _mm_setzero_si128();

  for (i = 0; i < ARRAYSIZE/16; i++) {
    next16numbers = _mm_loadu_si128 ((__m128i *) (numberarray+i*16));
    next16numbers = _mm_cmpgt_epi8(next16numbers, comparatorMask);
    accumulator   = _mm_add_epi8(accumulator, next16numbers);
    if((i+1)%128==0) {
      for(j=0; j<16; j++) {
        parallelcnt += (signed char) _mm_extract_epi8(accumulator,j);
      }

      accumulator   = _mm_setzero_si128();
    }
  }

  if((i+1)%128!=0) {
    for(j=0; j<16; j++) {
      parallelcnt += (signed char) _mm_extract_epi8(accumulator,j);
    }
  }

  gettimeofday (&end, NULL);

  printf("Parallel-16 (version I) : count = %d\n",-parallelcnt);

#elif( MODE == 3 )

	/* Parallel count : 16 values at a time - version 2 */

  gettimeofday (&start, NULL);

  uint32_t parallelcnt = 0;
  __m128i next16numbers;
  const __m128i comparatorMask = _mm_set_epi8(42,42,42,42,42,42,42,42,42,42,42,42,42,42,42,42);
  __m128i zero                 = _mm_setzero_si128();
  __m128i accumulator          = _mm_setzero_si128();

  for (int i = 0; i < ARRAYSIZE/16; i++) {
    next16numbers = _mm_loadu_si128 ((__m128i *) (numberarray+i*16));
	  next16numbers = _mm_cmpgt_epi8(next16numbers, comparatorMask);
	  next16numbers = _mm_abs_epi8(next16numbers);
	  next16numbers = _mm_sad_epu8(next16numbers, zero);
	  accumulator   = _mm_add_epi64(accumulator, next16numbers);
  }
  parallelcnt   = _mm_extract_epi64(accumulator,0) + _mm_extract_epi64(accumulator,1);

  gettimeofday (&end, NULL);

  printf("Parallel-16 (version II) : count = %d\n",parallelcnt);

#elif( MODE == 4 )

  gettimeofday (&start, NULL);

  uint32_t serialcnt = 0;
  for (i = 0; i < ARRAYSIZE; i++) {
      serialcnt += (numberarray[i] > 42);
  }

  gettimeofday (&end, NULL);
  printf("Serial (predicated) : count = %d\n",serialcnt);

#else

	/* Serial count */

  gettimeofday (&start, NULL);

  uint32_t serialcnt = 0;
  for (i = 0; i < ARRAYSIZE; i++) {
    if(numberarray[i] > 42) {
      serialcnt++;
    }
  }

  gettimeofday (&end, NULL);
  printf("Serial : count = %d\n",serialcnt);

#endif

  time1 = (end.tv_sec - start.tv_sec) * 1000000 + (end.tv_usec - start.tv_usec);
  printf ("Required %lu usec to count.\n\n", time1);

  free(numberarray);

  return 0;
}
